#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <netinet/in.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>
#include <string.h>
#include <signal.h>
#include <sys/socket.h>
#include <net/if.h>
#include <stdio.h>
#include <sys/uio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include "net_pcap.h"
#include "global.h"
#include "md5.h"
#include "plc.h"

#ifndef SOL_PACKET
#warning SOL_PACKET undeclared !
#define SOL_PACKET 263
#endif

/*
 * Dans la trame Channel capabilities, on ne peut pas savoir  si les donnees
 * concernent le TX ou le RX. Il faut donc se rappeler de la derniere commande
 * envoyee, et en plus de la derniere commande envoyee par les autres...
 */
unsigned char capabilities_flag = 0;
static pcap_t *handle = NULL;
static QString if_mac;
static pcap_handler cb;
static char filter_exp[] = "ether[12]==0x88 && ether[13] == 0x7b";

unsigned short ex_word(const unsigned char *ptr)
{
	return ntohs(*((unsigned short*)ptr));
}

unsigned long ex_long(const unsigned char *ptr)
{
	return ntohl(*((unsigned long*)ptr));
}

char *format_mac_addr(const unsigned char *addr, char *macbuf)
{
	sprintf(macbuf, "%02x:%02x:%02x:%02x:%02x:%02x", addr[0], addr[1], addr[2], addr[3], addr[4], addr[5]);
	return macbuf;
}

int plcnet_select_if(const QNetworkInterface plcnet_if)
{
	char dev[50]={0};
	char errbuf[PCAP_ERRBUF_SIZE];
	/*	Ethernet Header is : Dest Addr (6 bytes) + Src Addr (6 bytes) + ether type (2 bytes) */
	bpf_u_int32 mask;
	bpf_u_int32 net;
	struct bpf_program fp;
	int optimize = 1;
	const QString if_name = plcnet_if.name();
	if_mac  = plcnet_if.hardwareAddress();

	strncpy(dev, if_name.toAscii().constData(), if_name.size());

	qDebug() << "plcnet_add_if with pcap";
	qDebug() << "plcnet_add_if with pcap : " << if_name;
	if (handle != NULL)
	{
		plcnet_deinit();
		sleep(1);
	}
	if (pcap_lookupnet(dev, &net, &mask, errbuf) == -1) {
		qDebug("Can't get netmask for device %s", dev);
		net = 0;
		mask = 0;
		return -1;
	}
	handle = pcap_open_live(dev, BUFSIZ, 1, 1000, errbuf);
	if (handle == NULL) {
		qDebug() << "Couldn't open device" << if_name << " error: " << errbuf;
		return -1;
	}
	if (pcap_compile(handle, &fp, filter_exp, optimize, mask) == -1)
	{
		qDebug() << "Couldn't parse filter" << filter_exp << pcap_geterr(handle);
		return -1;
	}
	if (pcap_setfilter(handle, &fp) == -1) {
		qDebug() << "Couldn't install filter" << filter_exp << pcap_geterr(handle);
		return -1;
	}
	qDebug("plcnet_add_if with pcap (fini)");
	return 0;
}

int plcnet_send_req(const void *buf, size_t len, const QString to_mac)
{
	unsigned char buffer[len+14];
	int i;

	if (handle == NULL)
		return -1;

	// Write dest
	std::wistringstream Stream(to_mac.toStdWString());
	for (i=0;i < ETHER_ADDR_LEN; i++)
	{
		unsigned Byte;
		wchar_t Sep;
		if (i > 0)
		{
			Stream >> Sep;
		}
		Stream >> std::hex >> Byte;
		buffer[i] = Byte;
	}

	// write source
	std::wistringstream Stream2(if_mac.toStdWString());
	for (i=0;i < ETHER_ADDR_LEN; i++)
	{
		unsigned Byte;
		wchar_t Sep;
		if (i > 0)
		{
			Stream2 >> Sep;
		}
		Stream2 >> std::hex >> Byte;
		buffer[ETHER_ADDR_LEN+i] = Byte;
	}
	// write ether typr
	buffer[ETHER_ADDR_LEN*2  ] = ETHERTYPE_INTELLON >> 8;
	buffer[ETHER_ADDR_LEN*2+1] = ETHERTYPE_INTELLON & 0xFF;
	// write data
	memcpy(buffer + 2 + (2 * ETHER_ADDR_LEN), buf, len);
	return pcap_sendpacket(handle, (const unsigned char*)buffer, len+14);
}


int plcnet_register_callback(pcap_handler callback, u_char *args)
{
	cb = callback;

	if (handle == NULL)
	{
		qDebug("Callback can not be registered, handle is NULL !");
		return -1;
	}
#define SNIFF_INFINITE -1
	return pcap_loop(handle, SNIFF_INFINITE, &plc_net_callback, args);
}

// Filtrer incoming packets, only Intellon types packets goes through
void plc_net_callback(u_char *args, const struct pcap_pkthdr *header, const u_char *packet)
{
	const struct sniff_ethernet *ethernet; /* The ethernet header */

	if (handle == NULL)
		return;

	ethernet = (struct sniff_ethernet*)(packet);
	if (ntohs(ethernet->ether_type) == ETHERTYPE_INTELLON)
	{
		cb(args, header, packet);
	}
	else
	{
		qWarning("received Ethernet frame of %X type, pcap filter is dead?", ntohs(ethernet->ether_type));
	}
}

#if 0
int plc_req_param_stat()
{
	unsigned char mme[3];
	qDebug("send : request parameters & statistics\n");
	mme[0] = 0x01; /* one MAC management entry */
	mme[1] = 0x07;	/* request parameters and statistics */
	mme[2] = 0x0;  /* 0 bytes follow */
	plcnet_send_req(mme, sizeof(mme), "ff:ff:ff:ff:ff:ff");
	return 0;
}
#endif

int plc_set_network_key(const QString& mac, const QString& key)
{
	unsigned char mme[12];
	// Use PBKDF1 to convert plaintext password
	int i;
	MD5 myMD5;
	md5_byte_t digest[16];

	/* generate initial digest */
	myMD5.append((const md5_byte_t *)key.toAscii().constData(), key.length());
	myMD5.append((const md5_byte_t *)"\x08\x85\x6d\xaf\x7c\xf5\x81\x85", 8);
	myMD5.finish();
	memcpy(digest, myMD5.getDigest(), 16);

	/* loop 999 times as required by HomePlug */
	for (i = 0; i < 999; i++) {
		myMD5.reset();
		myMD5.append(digest, 16);
		myMD5.finish();
		memcpy(digest, myMD5.getDigest(), 16);
	}
	/* extract the first 8 bytes; calculate parity bit (LSB = odd parity) */
	for (i = 0; i < 8; i++)
	{
		digest[i] = deskeyparity(digest[i]);
	}

	/* MMECTRL   = 01
	 * MEHDR = (MEV<<5) + MTYPE  = 0x04
	 * MMELength = 09
	 * EKS    = 01 (encryption key select)
	 * NEK (8 bytes)
	 */
	qDebug() << "send : set network key to :" << key;
	mme[0] = 0x01; /* one MAC management entry */
	mme[1] = 0x04;	/* set network key */
	mme[2] = 0x09;	/* 9 bytes follow */
	mme[3] = 0x01;	/* encryption key select -> 1 */
	qDebug("NEK = ");
	for (i = 0; i < 8; i++)
	{
		mme[i+4] = digest[i];
		qDebug("0x%X ", digest[i]);
	}
	plcnet_send_req(mme, sizeof(mme), mac);
	return 0;
}

#if 0
int plc_req_intellon_specific()
{
	int i;

	qDebug("send : request Intellon-specific network statistics\n");
	outframe[0] = 0x01; /* one MAC management entry */
	outframe[1] = 0x1a;	/* request network statistics */
	outframe[2] = 0xbb;	/* 187 bytes follow */
	outframe[3] = 0x80;	/* read the stats, don't clear them */
	for (i = 0; i < 186; i++)
	{
		outframe[i+4] = 0x00;
	}
	sendto(netfd, outframe, 190, 0, (struct sockaddr *)&addr_broadcast, sizeof(addr_broadcast));
	return 0;
}
#endif

#if 0
int plc_req_perf_stat()
{
	int i;
	qDebug("send : request perf statistics");
	outframe[0] = 0x01; /* one MAC management entry */
	outframe[1] = 0x1c;	/* request network statistics */
	outframe[2] = 123;	/* 123 bytes follow */
	outframe[3] = 0x80;	/* read the stats, don't clear them */
	for (i = 0; i < 122; i++)
	{
		outframe[i+4] = 0x00;
	}
	return sendto(netfd, outframe, 190, 0, (struct sockaddr *)&addr_broadcast, sizeof(addr_broadcast));
}
#endif

int plcnet_deinit()
{
	qDebug() << "plcnet_deinit";

	if (handle !=NULL)
	{
		qDebug() << "pcap_close";
		pcap_close(handle);
		handle = NULL;
	}
	return 0;
}

// send mac management
int plc_send_mm()
{
	unsigned char mme[] = { 0x01, 0x02, 0x05, 0x00, 0x04, 0x87, 0x10};

	/*   
	 * 	MMECTRL   = 01
	 * 	MMEHeader = 02
	 * 	MMELength = 05
	 *	MMEOUI    = 00 04 87
	 *      Ox10 : Bit 7 = 0 (Request)
	 *           Bits 6-0, MID = 0x10 (16 = Get Device Description)
	 */
	qDebug("plc_send_mm : Get Device Description");
	return plcnet_send_req(mme, sizeof(mme), "ff:ff:ff:ff:ff:ff");
}


int plc_broadcast_key()
{
	unsigned char mme[] = { 0x01, 0x14, 0x0A, 0x01, 0x00,
		0x9b, 0x1a, 0xa2, 0x0d, 0x83, 0xcb, 0x6e, 0xc8}; // always same data !! Tell me if you find what to put here ! TODO !

	/* 	MMECTRL   = 01
	 * 	MEHDR = (MEV<<5) + MTYPE  = 0x14
	 * 	MMELength = 0x0A
	 *	MID       = 0x01 Encryption Key Tag Message ID
	 *	EKS       = 0x00 Encryption Key Select
	 */
	qDebug("plc_broadcast_key");
	return plcnet_send_req(mme, sizeof(mme), "ff:ff:ff:ff:ff:ff");
}


#if 0
int plc_get_device_version(const QString &mac)
{
	unsigned char mme[] = { 0x01, 0x12, 0x05, 0x00, 0x04, 0x87, 0x1};
	struct sockaddr_ll addr= { 0,0,0,0,0,0,{0,} };
	int i;

	/*  
	 *    MMECTRL   = 01
	 * 	MEHDR = (MEV<<5) + MTYPE  = 0x12
	 *    MMELength = 05
	 *    MMEOUI    = 00 04 87
	 *    Ox10 : Bit 7 = 0 (Request)
	 *           Bits 6-0, MID = 0x1 (1 = Get Device Version)
	 */
	addr.sll_family  = AF_PACKET;
	addr.sll_protocol= htons(ETHERTYPE_INTELLON);
	addr.sll_ifindex = ifr.ifr_ifindex;

	std::wistringstream Stream(mac.toStdWString());

	for (i=0;i < 6; i++)
	{
		unsigned Byte;
		wchar_t Sep;
		if (i > 0)
		{
			Stream >> Sep;
		}
		Stream >> std::hex >> Byte;
		addr.sll_addr[i] = Byte;
	}

	qDebug("plc_get_device_version ");
	for (i=0; i < 6; ++i)
	{
		qDebug("<%02x>",addr.sll_addr[i]);
	}
	qDebug("Get Device Version");
	return plcnet_send_req(mme, sizeof(mme), mac);
}
#endif

int plc_get_channel_capacities()
{
	int res1, res2;
	unsigned char mme[] = { 0x01, 0x02, 0x06, 0x00, 0x04, 0x87, 0x18, 0x0};
	/*
	 * 	MMECTRL   = 01
	 * 	MMEHeader = 02
	 * 	MMELength = 05
	 *	MMEOUI    = 00 04 87
	 *      Ox18 : Bit 7 = 0 (Request)
	 *           Bits 6-0, MID = 0x18 (24 = Get Channel Capacities)
	 *      0x00 : (0 : Request Receive Capacities, 01 : transmit)
	 *
	 */
	qDebug("plc_send_mm : Get Channel capacities for receive");
	capabilities_flag = CAPABILITIES_RX;
	res1 = plcnet_send_req(mme, sizeof(mme), "ff:ff:ff:ff:ff:ff");
	sleep(1);
	mme[7] = 1; // Request transmit Capacities
	qDebug("plc_send_mm : Get Channel capacities for transmit");
	capabilities_flag = CAPABILITIES_TX;
	res2 = plcnet_send_req(mme, sizeof(mme), "ff:ff:ff:ff:ff:ff");
	return ((res1>=0) && (res2>=0));
}

#if 0
int plc_req_channel_estimation()
{
	unsigned char mme[] = { 0x01, 0x00, 0x01, 0x00};
	struct sockaddr_ll addr= { 0,0,0,0,0,0,{0,} };
	/*
	 * 	MCTRL = NE     = 01
	 * 	MEHDR = (MEV<<5) + MTYPE  = 00 : Request Channel Estimation
	 * 	MELen = 01
	 * 	MMENTRY = CEV       = 0 : INT5500
	 */
	addr.sll_family  = AF_PACKET;
	addr.sll_protocol= htons(ETHERTYPE_INTELLON);
	addr.sll_ifindex = ifr.ifr_ifindex;
	addr.sll_addr[0] = 0x00;
	addr.sll_addr[1] = 0x0C;
	addr.sll_addr[2] = 0xb9;
	addr.sll_addr[3] = 0x03;
	addr.sll_addr[4] = 0x6f;
	addr.sll_addr[5] = 0xc1;

	qDebug() << "plc_req_channel_estimation to" << mac;
	return plcnet_send_req(mme, sizeof(mme), "ff:ff:ff:ff:ff:ff");
}
#endif

// read Firmware Version
int plc_req_version(const QString &mac)
{
	unsigned char req[258] = {                                                                    0x01, 0x12,
		0x01, 0x20, 0x00, 0x00, 0xe5, 0x03, 0x00, 0x00,   0x68, 0xfb, 0x12, 0x00, 0x82, 0x93, 0x80, 0x7c,
		0xe5, 0x03, 0x00, 0x00, 0xb4, 0x88, 0x44, 0x00,   0xa8, 0xfb, 0x12, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x16, 0x0e, 0x81, 0x7c, 0xf8, 0x00, 0x00, 0x00,   0xc8, 0x88, 0x44, 0x00, 0x3c, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,   0xd0, 0xfb, 0x12, 0x00, 0x18, 0x1a, 0x80, 0x7c,
		0x74, 0xfb, 0x12, 0x00, 0x4c, 0x1f, 0x43, 0x00,   0xf8, 0xfc, 0x12, 0x00, 0xa8, 0x9a, 0x83, 0x7c,
		0x20, 0x0e, 0x81, 0x7c, 0xff, 0xff, 0xff, 0xff,   0x16, 0x0e, 0x81, 0x7c, 0xce, 0x4f, 0x00, 0x10,
		0x78, 0x00, 0x00, 0x00, 0xc8, 0x88, 0x44, 0x00,   0x3c, 0x00, 0x00, 0x00, 0xf4, 0xfb, 0x12, 0x00,
		0x00, 0xf0, 0xfd, 0x7f, 0x1c, 0xfc, 0x12, 0x00,   0x3c, 0x24, 0x80, 0x7c, 0xe0, 0xfb, 0x12, 0x00,
		0x0f, 0x24, 0x80, 0x7c, 0x9c, 0x92, 0x80, 0x7c,   0x77, 0x55, 0xb0, 0x01, 0x00, 0x00, 0x00, 0x00,
		0x14, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00,   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x10, 0x00, 0x00, 0x00, 0xf0, 0xd8, 0xff, 0xff,   0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00,
		0xf4, 0xfb, 0x12, 0x00, 0xd4, 0xfb, 0x12, 0x00,   0x60, 0x7d, 0x41, 0x00, 0xf8, 0xfc, 0x12, 0x00,
		0xa8, 0x9a, 0x83, 0x7c, 0x58, 0x24, 0x80, 0x7c,   0xff, 0xff, 0xff, 0xff, 0x0f, 0x24, 0x80, 0x7c,
		0x51, 0x24, 0x80, 0x7c, 0x01, 0x00, 0x00, 0x00,   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x76, 0xa4, 0x40, 0x00, 0x7d, 0xa4, 0x40, 0x00,   0x00, 0x00, 0x00, 0x00, 0x60, 0x7d, 0x41, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,   0x03, 0xa0, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00
	};
	qDebug() << "plc_req_version to " << mac;
	return plcnet_send_req(req, sizeof(req), mac);
}

